#include <stdio.h>
#include <string.h>
#include "mpi.h"

#define REP 5

/* Variables globales */
double t1, t2, tmin, tmax, t;
int rank, p;
int i, j, k;
int ldata;
int *data, *datap;
int tag;
int master=0;
MPI_Status status;
MPI_Request request;

/* Imprime tiempos */
void print_time(char *msg) {
	MPI_Reduce(&t1, &tmin, 1, MPI_DOUBLE, MPI_MIN, master, MPI_COMM_WORLD);
	MPI_Reduce(&t2, &tmax, 1, MPI_DOUBLE, MPI_MAX, master, MPI_COMM_WORLD);
	t = (tmax - tmin);
	if (rank == master)
		printf("[%s] P=%d, N=%d, S=%.8lf\n", msg, p, ldata, t);
}

/* Envío y recepción asíncrono */
void send_recv() {
	tag = 0;
	t1 = MPI_Wtime();
	if (rank == master)  /* Envío a todos, menos el master */
		for (i=0; i<p; i++) {
			if (i != master)
				MPI_Send(data, ldata, MPI_INT, i, tag, MPI_COMM_WORLD);
		}
	else /* Recepción de un mensaje desde el master */
		MPI_Recv(data, ldata, MPI_INT, master, tag, MPI_COMM_WORLD, &status);
	
	t2 = MPI_Wtime();
	fprintf(stderr, "(%d/%d) %s, T1=%.8lf, T2=%.8lf\n", 
		rank, p, rank==master ? "MPI_Send":"MPI_Recv", t1, t2);
	print_time("MPI_Send/MPI_Recv");
}

void isend_irecv() {
	tag = 1;
        t1 = MPI_Wtime();
        if (rank == master)  /* Envío a todos, menos el master */
                for (i=0; i<p; i++) {
                        if (i != master)
                                MPI_Isend(data, ldata, MPI_INT, i, tag, MPI_COMM_WORLD, &request);
                }
        else { /* Recepción de un mensaje desde el master */
                MPI_Irecv(data, ldata, MPI_INT, master, tag, MPI_COMM_WORLD, &request);
		MPI_Wait(&request, &status);
	}

        t2 = MPI_Wtime();
        fprintf(stderr, "(%d/%d) %s, T1=%.8lf, T2=%.8lf\n", 
		rank, p, rank==master ? "MPI_Isend":"MPI_Irecv", t1, t2);
	print_time("MPI_Isend/MPI_Irecv");
}

void broadcast() {
	tag = 2;

	t1 = MPI_Wtime();
	MPI_Bcast(data, ldata, MPI_INT, master, MPI_COMM_WORLD);
	t2 = MPI_Wtime();
	fprintf(stderr, "(%d/%d) MPI_Bcast, T1=%.8lf, T2=%.8lf\n", rank, p, t1, t2);
	print_time("MPI_Bcast");
}

void gather() {
	tag = 3;

	t1 = MPI_Wtime();
	MPI_Gather(data, ldata, MPI_INT, datap, ldata, MPI_INT, master, MPI_COMM_WORLD);
	t2 = MPI_Wtime();
	fprintf(stderr, "(%d/%d) MPI_Gather, T1=%.8lf, T2=%.8lf\n", rank, p, t1, t2);
	print_time("MPI_Gather");
}

void scatter() {
        tag = 5;

	t1 = MPI_Wtime();
	MPI_Scatter(datap, ldata, MPI_INT, data, ldata, MPI_INT, master, MPI_COMM_WORLD);
	t2 = MPI_Wtime();
	fprintf(stderr, "(%d/%d) MPI_Scatter, T1=%.8lf, T2=%.8lf\n", rank, p, t1, t2);
	print_time("MPI_Scatter");
}

void all() {
	fprintf(stderr, "****** PRIMITIVA: MPI_Send/MPI_Recv *********");
	send_recv();
	fprintf(stderr, "****** PRIMITIVA: MPI_Isend/MPI_Irecv *******");
	isend_irecv();
	fprintf(stderr, "****** PRIMITIVA: MPI_Bcast *****************");
	broadcast();
	fprintf(stderr, "****** PRIMITIVA: MPI_Gather ****************");
	gather();
	fprintf(stderr, "****** PRIMITIVA: MPI_Scatter ***************");
	scatter();
}

void gm() {
	if (!(datap = malloc((ldata+1)*p*sizeof(int)))) {
		fprintf(stderr, "(%d/%d) No hay memoria para malloc secundario\n", rank, p);
		exit(1);
	}
}

void fm() {
	free(datap);
}

/* MAIN */
int main(int argc, char **argv) {
	char option;

	/* Lee parámetros */
	if (argc != 3) {
		printf("Uso: %s prueba N\n", argv[0]);
		printf("\tprueba\ta (all), s (sincrono), i (asincrono), b (broadcast), g (gather), c (scatter)\n");
		printf("\tN\tBytes del mensaje\n");
		exit(1);
	}
	option = argv[1][0];
	ldata = atoi(argv[2]);

	/* Inicializa MPI */
	MPI_Init(&argc, &argv);

	/* Obtiene número de proceso en rank */
	MPI_Comm_rank(MPI_COMM_WORLD, &rank);
	
	/* Número de procesos en el comunicador */
	MPI_Comm_size(MPI_COMM_WORLD, &p);

	/* Reserva memoria */
	if (!(data = malloc((ldata+1)*sizeof(int)))) {
		fprintf(stderr, "(%d/%d) No hay memoria para malloc principal\n", rank, p);
		exit(1);
	}

	/* Ejecuta rutinas */
	switch(option) {
		case 'a': all(); break;
		case 's': send_recv(); break;
		case 'i': isend_irecv(); break;
		case 'b': broadcast(); break;
		case 'g': gm(); gather(); fm(); break;
		case 'c': gm(); scatter(); fm(); break;
		default: printf("Prueba no válida\n");
	}
	
	/* Finalizar */
	free(data);
	MPI_Finalize();
	exit(0);
}
